#ifndef CAMERA_H
#define CAMERA_H

typedef enum {
	PERSPECTIVE_CAMERA,
	ORTHOGRAPHIC_CAMERA
} CameraType;

typedef struct {
	float	fovy;
	float	near;
	float	far;
} PerspCamera;

typedef struct {
	float	left;
	float	right;
	float	bottom;
	float	top;
	float	near;
	float	far;
} OrthoCamera;

typedef struct {
	CameraType type;
	union {
		PerspCamera persp;
		OrthoCamera ortho;
	} cam;
} Camera;

void init_Camera( Camera* camera, CameraType type );
void destroy_Camera( Camera* camera );

void setPerspParams_Camera( Camera* camera, float fovy, float near, float far );
float getPerspFOVY_Camera( const Camera* camera );
float getPerspNear_Camera( const Camera* camera );
float getPerspFar_Camera( const Camera* camera );

void setOrthoParams_Camera( Camera* camera, float left, float right,
	float bottom, float top, float near, float far );
float getOrthoLeft_Camera( const Camera* camera );
float getOrthoRight_Camera( const Camera* camera );
float getOrthoBottom_Camera( const Camera* camera );
float getOrthoTop_Camera( const Camera* camera );
float getOrthoNear_Camera( const Camera* camera );
float getOrthoFar_Camera( const Camera* camera );

#endif	/* CAMERA_H */

