#include <abort.h>
#include <malloc.h>
#include <string.h>
#include <containers/vec_.h>

Vec* new_Vec( size_t el_size, size_t alignment ) {
	Vec* vec = malloc( sizeof(Vec) );
	if(NULL == vec)
		abort();
	init_Vec( vec, el_size, alignment );
	return vec;
}

inline void delete_Vec( Vec* self ) {
	destroy_Vec( self );
	free( self );
}

void init_Vec( Vec* self, size_t el_size, size_t alignment ) {
	self->el_size = el_size;
	self->alignment = alignment;
	self->num_el = 0;
	self->max_el = 1;
	if( NULL == (
		self->data_unaligned =
		 	malloc( el_size*self->max_el + (1<<alignment) )
		)
	) {
		abort();
	}
	self->end = self->data = ALIGN_ADR( self->data_unaligned, alignment );
}

inline void destroy_Vec( Vec* self ) {
	free( self->data );
}

void resize_Vec( Vec* self, uint32_t newsize ) {
	if( NULL == (
		self->data_unaligned =
			realloc( self->data_unaligned, self->el_size*newsize + (1<<self->alignment) )
		)
	) {
		self->data = NULL;
		self->end = NULL;
		abort();
	}
	self->data = ALIGN_ADR( self->data_unaligned, self->alignment );
	self->end = (uint8_t*)self->data + self->el_size * self->num_el;
	self->max_el = newsize;
}

void pushBack_Vec( Vec* self, const void* element ) {
	if( self->num_el >= self->max_el ) {
		resize_Vec( self, self->max_el + 16 );	// it used to be vec->max_el << 1, but seems like it triggers a BUG!
	}

	memcpy( self->end, element, self->el_size );
	self->end = (uint8_t*)self->end + self->el_size;
	self->num_el++;
}

inline void* getElement_Vec( Vec* self, uint32_t i ) {
	return (uint8_t*)self->data + i * self->el_size;
}

inline const void* getConstElement_Vec( const Vec* self, uint32_t i ) {
	return (uint8_t*)self->data + i * self->el_size;
}

inline void* getTail_Vec( Vec* self ) {
	return (uint8_t*)self->data + self->num_el * self->el_size;
}


inline uint32_t getLength_Vec( const Vec* self ) {
	return self->num_el;
}

inline void reset_Vec( Vec* self ) {
	self->num_el = 0;
	self->end = self->data;
}

