#include <sys/region.h>
#include <node/node.h>

void nodeConstruct(Node* node) {
	init_Vec(&node->children, sizeof(Node*), ALIGN_DEFAULT);
	node->type = NODE_RAW;
	node->parent = NULL;
}

void nodeDestroy(Node* node) {
	destroy_Vec(&node->children);
}

void nodeAttach(Node* node, Node* child) {
	pushBack_Vec(&node->children, &child);
	child->parent = node;
}

inline uint_t nodeGetNumChildren(const Node* node) {
	return getLength_Vec(&node->children);
}

inline Node* nodeGetParent(Node* node) {
	return node->parent;
}

inline Node* nodeGetChild(Node* node, uint_t i) {
	return *(Node**)getElement_Vec(&node->children, i);
}

inline void nodeIteratorConstruct(NodeIterator* iterator) {
	init_Stack(&iterator->stack, sizeof(iterator->i));
	nodeIteratorReset(iterator, NULL);
}

inline void nodeIteratorDestroy(NodeIterator* iterator) {
	destroy_Stack(&iterator->stack);
}

Node* nodeIteratorInc(NodeIterator* iterator) {
	Node* node;
i_believe_in_goto:	node = iterator->next;
	if(NULL == node)
		return NULL;

	bool node_visited = iterator->i;

	uint_t n = nodeGetNumChildren(node);
	if(iterator->i < n) {

		iterator->next = nodeGetChild(node, iterator->i);
		iterator->i++;
		push_Stack(&iterator->stack, &iterator->i);

		iterator->i = 0;

	} else {

		if(true == isEmpty_Stack(&iterator->stack)) {
			if(0 != n)
				return NULL;

			/* it was a single-node tree */
			iterator->next = NULL;
			return iterator->node;
		}

		iterator->i = *(uint_t*)pop_Stack(&iterator->stack);
		iterator->next = nodeGetParent(node);
	}

	if(node_visited)
		goto i_believe_in_goto;

	return node;
}

inline void nodeIteratorReset(NodeIterator* iterator, Node* node) {
	iterator->node = node;
	iterator->next = iterator->node;
	iterator->i = 0;
}

