#include <abort.h>
#include "string.h"
#include "malloc.h"

#define HEAPS_NUM			17
#define HEAP16_ENTRIES		(2)
#define HEAP16_SIZE			((1<<16)*HEAP16_ENTRIES)	// 64k
#define HEAP15_ENTRIES		(2)
#define HEAP15_SIZE			((1<<15)*HEAP15_ENTRIES)				// 32k
#define HEAP14_ENTRIES		(2)
#define HEAP14_SIZE			((1<<14)*HEAP14_ENTRIES)				// 16k
#define HEAP13_ENTRIES		(25)
#define HEAP13_SIZE			((1<<13)*HEAP13_ENTRIES)			// 8k
#define HEAP12_ENTRIES		(50)
#define HEAP12_SIZE			((1<<12)*HEAP12_ENTRIES)			// 4k
#define HEAP11_ENTRIES		(50)
#define HEAP11_SIZE			((1<<11)*HEAP11_ENTRIES)			// 2k
#define HEAP10_ENTRIES		(50)
#define HEAP10_SIZE			((1<<10)*HEAP10_ENTRIES)			// 1k
#define HEAP9_ENTRIES		(50)
#define HEAP9_SIZE			((1<<9)*HEAP9_ENTRIES)				// 512
#define HEAP8_ENTRIES		(50)
#define HEAP8_SIZE			((1<<8)*HEAP8_ENTRIES)				// 256
#define HEAP7_ENTRIES		(50)
#define HEAP7_SIZE			((1<<7)*HEAP7_ENTRIES)				// 128
#define HEAP6_ENTRIES		(50)
#define HEAP6_SIZE			((1<<6)*HEAP6_ENTRIES)				// 64
#define HEAP5_ENTRIES		(50)
#define HEAP5_SIZE			((1<<5)*HEAP5_ENTRIES)				// 32
#define HEAP4_ENTRIES		(50)
#define HEAP4_SIZE			((1<<4)*HEAP4_ENTRIES)				// 16
#define HEAP3_ENTRIES		(50)
#define HEAP3_SIZE			((1<<3)*HEAP3_ENTRIES)				// 8
#define HEAP2_ENTRIES		(50)
#define HEAP2_SIZE			((1<<2)*HEAP2_ENTRIES)				// 4
#define HEAP1_ENTRIES		0
#define HEAP1_SIZE			0						// 2
#define HEAP0_ENTRIES		0
#define HEAP0_SIZE			0						// 1

#define HEAP_MAX_ALLOC_SIZE	(2<<16)
#define HEAP_MIN_ALLOC_SIZE	4
#define HEAP_MAX_ENTRIES	50
#define	HEAP_FREE_ENTRY		0xFFFFFFFF
#define ALIGNMENT_MASK		0x7	// 8 byte boundary

typedef struct {
	void*		base;
	uint32_t	numobj;
	uint32_t	objsize;
	uint32_t	ofs[HEAP_MAX_ENTRIES];
} Heap;

static Heap g_heaps[HEAPS_NUM];

void mallocInit( void* base, size_t size ) {
	static uint32_t heapsize[HEAPS_NUM] = {
		HEAP0_SIZE, HEAP1_SIZE, HEAP2_SIZE, HEAP3_SIZE,
		HEAP4_SIZE, HEAP5_SIZE, HEAP6_SIZE, HEAP7_SIZE,
		HEAP8_SIZE, HEAP9_SIZE, HEAP10_SIZE, HEAP11_SIZE,
		HEAP12_SIZE, HEAP13_SIZE, HEAP14_SIZE, HEAP15_SIZE,
		HEAP16_SIZE
	};
	static uint32_t heapentries[HEAPS_NUM] = {
		HEAP0_ENTRIES, HEAP1_ENTRIES, HEAP2_ENTRIES, HEAP3_ENTRIES,
		HEAP4_ENTRIES, HEAP5_ENTRIES, HEAP6_ENTRIES, HEAP7_ENTRIES,
		HEAP8_ENTRIES, HEAP9_ENTRIES, HEAP10_ENTRIES, HEAP11_ENTRIES,
		HEAP12_ENTRIES, HEAP13_ENTRIES, HEAP14_ENTRIES, HEAP15_ENTRIES,
		HEAP16_ENTRIES
	};

	void* curr_base = base;
	for( uint32_t j=0; j < HEAPS_NUM; curr_base = (void*)((uintptr_t)curr_base + heapsize[j++]) ) {
		Heap* heap = &g_heaps[j];
		heap->base = curr_base;
		heap->numobj = heapentries[j];
		heap->objsize = (1<<j);
		for( uint32_t i=0; i < HEAP_MAX_ENTRIES; i++ ) {
			heap->ofs[i] = HEAP_FREE_ENTRY;
		}
	}

	if( (uintptr_t)base + size < (uintptr_t)curr_base) {
		abort();
	}	// memory exceeded
}

void* realloc( void* ptr, size_t size ) {
	if( !ptr ) {
		return NULL;
	}
	
	static uint8_t buffer[HEAP_MAX_ALLOC_SIZE];

	memcpy( buffer, ptr, size );
	free( ptr );
	ptr = malloc( size );
	memcpy( ptr, buffer, size );
	
	return ptr;
}

#if 0
void* malloc( size_t size ) {
	if( size > HEAP_MAX_ALLOC_SIZE ) {
		return NULL;
	}
	if( size < HEAP_MIN_ALLOC_SIZE ) {
		size = HEAP_MIN_ALLOC_SIZE;
	}
	
	uint32_t mask = 0x80000000;
	uint32_t heapsel = 31;
	while( !(size & mask) ) {
		mask >>= 1;
		heapsel--;
	}
	if( size & (mask-1) ) {
		heapsel++;
	}
	
	Heap* heap = &g_heaps[heapsel];
	for( uint32_t i=0; i < heap->numobj; i++ ) {
		if( HEAP_FREE_ENTRY == heap->ofs[i] ) {
			heap->ofs[i] = heap->objsize*i;
			uintptr_t p = (uintptr_t)heap->base + heap->ofs[i];
			if( p & ALIGNMENT_MASK ) {
				abort();
			}
			return (void*)p;
		}
	}

	return NULL;
}

void free( void* ptr ) {
	for( uint32_t j=0; j < HEAPS_NUM; j++ ) {
		Heap* heap = &g_heaps[j];
		for(uint32_t i=0; i < heap->numobj; i++ ) {
			if( ((uintptr_t)heap->base + heap->ofs[i]) == (uintptr_t)ptr ) {
				heap->ofs[i] = HEAP_FREE_ENTRY;
				return;
			}
		}
	}
}
#else
extern uint32_t __heap_start;
extern uint32_t __heap_end;

typedef struct Header {
	struct Header* next;
	struct Header* prev;
	size_t block_size;
	uint32_t pad;		// we need this to keep addresses aligned
} Header;

static size_t calculate_spacing( Header* header ) {
	if( NULL == header->next ) { return 0; }

	return (uintptr_t)header->next - ( (uintptr_t)header + sizeof(Header) + (uintptr_t)header->block_size );
}

void* malloc( size_t size ) {
	size = (size + 0xF) & (~0xF);	// align to 16-bytes

	static bool initialized = false;

	static Header* head;
	static Header* tail;
	if( !initialized ) {
		head = (Header*)(0x80300000 + 320*240*2);	//__heap_start;
		tail = (Header*)(0x80800000 - 8192);		//((Header*)__heap_end) - 1;
	
		head->next = tail;
		head->prev = NULL;
		head->block_size = 0;

		tail->next = NULL;
		tail->prev = head;
		tail->block_size = 0;

		initialized = true;
	}

	Header* current = head;
	while( NULL != current->next ) {
		if( calculate_spacing( current ) >= size + sizeof(Header) ) {
			Header* new = (Header*)( (uintptr_t)current + sizeof(Header) + current->block_size );
			new->prev = current;
			new->next = current->next;
			new->block_size = size;

			current->next = new;
			current->next->prev = new;

			return (void*)((uintptr_t)new + sizeof(Header));
		}

		current = current->next;
	}

	return NULL;
}

void free( void* ptr ) {
	Header* header = (Header*)( (uintptr_t)ptr - sizeof(Header) );
	header->prev->next = header->next;
	header->next->prev = header->prev;
}

#endif
